import numpy as np

class TransformationMatrix:
    def __init__( self, value = None, ndim = None ):
        if isinstance( value, TransformationMatrix ):
            self.value = value.value
        elif value is None and ndim is not None:
            self.value = np.eye( ndim + 1 )
        else:
            self.value = value

    def scale( self, s ):
        if isinstance( s, float ) or isinstance( s, int ):
            self.value = self.get()
            for i in range( self.value.shape[ 0 ] ):
                self.value[ i, i ] *= s
        else:
            raise RuntimeError( "TODO" )

    def translate( self, v ):
        self.value = self.get( np.array( v ).size )
        self.value[ :-1, -1 ] = v

    def dir( self, d ):
        M = self.get( np.array( d ).size )
        return M[ :-1, :-1 ] @ d

    def pos( self, p ):
        n = np.array( list( p ) + [ 1 ] )
        M = self.get( np.array( p ).size )
        return M[ :-1, : ] @ n
    
    def get( self, ndim = None ):
        # None
        if self.value is None:
            return np.eye( ndim + 1 )

        # ( M, V )
        if isinstance( self.value, tuple ):
            assert( len( self.value ) == 2 )
            M = np.array( self.value[ 0 ] )
            V = np.array( self.value[ 1 ] )
            
            assert( M.shape[ 0 ] == M.shape[ 1 ] )
            assert( M.shape[ 0 ] == V.shape[ 0 ] )
            s = M.shape[ 0 ]

            res = np.eye( s + 1 )
            res[ :s, :s ] = M
            res[ :s, s ] = V
            return res
        
        # full matrix
        array = np.array( self.value )
        if len( array.shape ) == 2:
            assert( array.shape[ 0 ] == array.shape[ 1 ] )
            if ndim is not None:
                assert( array.shape[ 0 ] == ndim + 1 )
            return array
        
        # vector
        if len( array.shape ) == 1:
            s = array.size

            res = np.eye( s + 1 )
            res[ :s, s ] = array
            return res

        print( self.value )        
        raise ValueError( "invalid transformation" )



import numpy as np

class D2GTransportMap:
    """
        Discrete (diracs) to generic measure transport map (optimal or not)

        => properties return vectors (one value for each dirac)
    """
    def __init__( self, sdot_solver ):
        self.sdot_solver = sdot_solver

    @property
    def kantorovitch_potentials( self ):
        return self.sdot_solver.power_diagram.weights

    @property
    def brenier_potentials( self ):
        p = self.sdot_solver.dirac_positions
        return self.kantorovitch_potentials + np.linalg.norm( p )

    @property
    def barycenters( self ):
        return self.sdot_solver.power_diagram.cell_barycenters()



from .SdotPlan import SdotPlan

def optimal_transport_plan( source_measure, target_measure = None, norm = 2, stop_when = None, display = None, relaxation = 1 ):
    """ 
        optimal transport plan between source_distribution and target_distribution
    """
    sp = SdotPlan( source_measure, target_measure, norm, display = display )
    sp.default_relaxation = relaxation

    if stop_when and 'max_error_ratio' in stop_when: # TODO: integration in SdotSolver
        sp.max_error_target = stop_when[ 'max_error_ratio' ]

    sp.adjust_potentials()
    return sp





class CellFunction:
    """
        function that take coordinate and return the value of the cell_function with the corresponding cell instance

    """

    def __init__( self, power_diagram, cell_function ):
        self.power_diagram = power_diagram
        self.cell_function = cell_function

    def __call__( self, x ):
        cell = self.power_diagram.find_cell( x )
        if cell is None:
            return None
        return self.cell_function( cell, x )

    def plot( self, plt = None ):
        if self.power_diagram.ndim != 2:
            raise RuntimeError( "TODO: plot CellFunction for dim != 2" )
        
        if plt is None:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            fig = fig.add_subplot( projection = '3d' )
            self.plot( fig )
            plt.show()
            return

        # ( self, fig, color = 'black', linestyle = '-', linewidth = 2, low_dim_linewidth_coeff = 0.25, ray_length = 0.2, ray_color = None, ray_linestyle = '--', free_color = None, free_linestyle = ':', elevation_function = None ):

        #
        def on_cell( cell ):
            cell.plot_pyplot( plt, elevation_function = lambda x: self.cell_function( cell, x ) )
        self.power_diagram.for_each_cell( on_cell, 1 )



from .CellFunction import CellFunction
import numpy as np

class G2DTransportMap:
    """
        Generic measure to Discrete (dirac) transport map (optimal or not)

        => properties return function of $x$ (position in the space)
    """
    def __init__( self, sdot_solver ):
        self.sdot_solver = sdot_solver

    @property
    def kantorovitch_potential( self ):
        def cell_function( cell, x ):
            return cell.weight + np.linalg.norm( x - cell.seed_position )**2
        return CellFunction( self.sdot_solver.power_diagram, cell_function )

    @property
    def brenier_potential( self ):
        def cell_function( cell, x ):
            # return cell.weight / 2 + np.linalg.norm( x - cell.seed_position )**2 - np.linalg.norm( x )**2
            return np.dot( x, cell.seed_position ) - ( np.linalg.norm( cell.seed_position )**2 - cell.weight ) / 2
        return CellFunction( self.sdot_solver.power_diagram, cell_function )

    @property
    def barycenters( self ):
        def cell_function( cell, x ):
            return cell.seed_position
        return CellFunction( self.sdot_solver.power_diagram, cell_function )
